@@ -1,9 +1,9 @@
 /******************************************************************************
     (c) 2002 Matthew Fredette                 fredette@netbsd.org
- 
+
     Some modifications:
-    (c) 2003 Patrick Caulfield                 patrick@debian.org
-    
+    (c) 2003-2004 Christine Caulfield           christine.caulfield@googlemail.com
+
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
@@ -25,6 +25,7 @@
 #include <sys/time.h>
 #include <sys/socket.h>
 #include <sys/uio.h>
+#include <assert.h>
 
 #include <net/if.h>
 #include <net/bpf.h>
@@ -383,7 +384,7 @@ int BPFInterfaces::send_packet(int ifn, unsigned char macaddr[], unsigned char *
   /* write this packet: */
   iov[0].iov_base = (char* )&ether_packet;
   iov[0].iov_len = sizeof(ether_packet);
-  iov[1].iov_base = data;
+  iov[1].iov_base = (char *)data;
   iov[1].iov_len = len;
   if (writev(_latd_bpf_fd, iov, 2) < 0) {
     syslog(LOG_ERR, "writev: %m");
@@ -394,21 +395,39 @@ int BPFInterfaces::send_packet(int ifn, unsigned char macaddr[], unsigned char *
 }
 
 
+#define CONT_OR_RET() \
+    if (_latd_bpf_buffer_offset < _latd_bpf_buffer_end) {\
+        continue;\
+    } else {\
+        if (_latd_bpf_buffer_offset_next <= _latd_bpf_buffer_end) \
+          more = true; \
+        return 0; \
+    }
+
 // Receive a packet from a given interface
-int BPFInterfaces::recv_packet(int sockfd, int &ifn, unsigned char macaddr[], unsigned char *data, int maxlen)
+int BPFInterfaces::recv_packet(int sockfd, int &ifn, unsigned char macaddr[], unsigned char *data, int maxlen, bool &more)
 {
   ssize_t buffer_end;
   struct bpf_hdr the_bpf_header;
   unsigned int _latd_bpf_buffer_offset_next;
 
+  more = false;
+
   /* loop until we have something to return: */
   for(;;) {
 
-      debuglog(("PJC: buffer_offset=%d, buffer_end=%d\n", _latd_bpf_buffer_offset,_latd_bpf_buffer_end));
+      debuglog(("buffer_offset=%d, buffer_end=%d\n", _latd_bpf_buffer_offset,_latd_bpf_buffer_end));
     /* if the buffer is empty, fill it: */
     if (_latd_bpf_buffer_offset
 	>= _latd_bpf_buffer_end) {
 
+	// Is there really anything to read? we really don't want to
+	// block here.
+	int bytes;
+	if (ioctl(sockfd, FIONREAD, &bytes) == 0 &&
+	    bytes == 0) {
+	    return 0;
+	}
       /* read the BPF socket: */
       debuglog(("bpf: calling read\n"));
       buffer_end = read(sockfd,
@@ -450,7 +469,7 @@ int BPFInterfaces::recv_packet(int sockfd, int &ifn, unsigned char macaddr[], un
 		the_bpf_header.bh_caplen, the_bpf_header.bh_datalen,
 		_latd_bpf_buffer_offset, _latd_bpf_buffer_end));
       _latd_bpf_buffer_offset = _latd_bpf_buffer_offset_next;
-      return 0;
+      CONT_OR_RET();
     }
 
     /* silently ignore packets that don't even have Ethernet headers,
@@ -462,8 +481,12 @@ int BPFInterfaces::recv_packet(int sockfd, int &ifn, unsigned char macaddr[], un
 		   _latd_bpf_interface_addr,
 		   ETHER_ADDR_LEN)) {
       /* silently ignore packets from us: */
+	if (the_bpf_header.bh_datalen < sizeof(struct ether_header))
+	    debuglog(("ignoring short packet of %d bytes\n", the_bpf_header.bh_datalen));
+	else
+	    debuglog(("ignoring packet from us\n"));
       _latd_bpf_buffer_offset = _latd_bpf_buffer_offset_next;
-      return 0;
+      CONT_OR_RET();
     }
     debuglog(("bpf: packet from %02x:%02x:%02x:%02x:%02x:%02x\n",
 	      ((struct ether_header *) (_latd_bpf_buffer + _latd_bpf_buffer_offset))->ether_shost[0],
@@ -495,18 +518,37 @@ int BPFInterfaces::recv_packet(int sockfd, int &ifn, unsigned char macaddr[], un
 				     + _latd_bpf_buffer_offset))->ether_shost,
 	   ETHER_ADDR_LEN);
     _latd_bpf_buffer_offset = _latd_bpf_buffer_offset_next;
+    if (_latd_bpf_buffer_offset < _latd_bpf_buffer_end)
+	more = true;
     return (the_bpf_header.bh_datalen
 	    - sizeof(struct ether_header));
   }
   /* NOTREACHED */
 }
 
+/* This is the LAT multicast address */
+static void store_lat_multicast(unsigned char *addr)
+{
+    addr[0]  = 0x09;
+    addr[1]  = 0x00;
+    addr[2]  = 0x2b;
+    addr[3]  = 0x00;
+    addr[4]  = 0x00;
+    addr[5]  = 0x0f;
+}
+
+
 // Open a connection on an interface
 int BPFInterfaces::set_lat_multicast(int ifn)
 {
     struct ifreq interface_ifreq;
     struct bpf_program program;
     int dummy_fd;
+#ifdef HAVE_AF_LINK
+    struct sockaddr_dl *sockdl;
+#endif
+
+    memset(&interface_ifreq, 0, sizeof(interface_ifreq));
 
     /* if we don't have an interface, bail: */
     if (ifn != 0 || _latd_bpf_interface_name == NULL) {
@@ -554,17 +596,20 @@ int BPFInterfaces::set_lat_multicast(int ifn)
       return -1;
     }
 
-    /* This is the LAT multicast address */
+#ifdef HAVE_AF_LINK
+    sockdl = (sockaddr_dl *)&interface_ifreq.ifr_addr;
+    sockdl->sdl_family = AF_LINK;
+    sockdl->sdl_alen = 6;
+    store_lat_multicast((unsigned char *)sockdl->sdl_data);
+#else
     interface_ifreq.ifr_addr.sa_family = AF_UNSPEC;
+    store_lat_multicast((unsigned char *)interface_ifreq.ifr_addr.sa_data);
+#endif
 #ifdef HAVE_SOCKADDR_SA_LEN
     interface_ifreq.ifr_addr.sa_len = sizeof(interface_ifreq.ifr_addr);
 #endif /* HAVE_SOCKADDR_SA_LEN */
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[0]  = 0x09;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[1]  = 0x00;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[2]  = 0x2b;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[3]  = 0x00;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[4]  = 0x00;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[5]  = 0x0f;
+
+
     if (ioctl(dummy_fd, SIOCADDMULTI, &interface_ifreq) < 0) {
       debuglog(("bpf: failed to add to the multicast list for interface for %s: %s\n",
 		interface_ifreq.ifr_name, strerror(errno)));
@@ -584,6 +629,11 @@ int BPFInterfaces::remove_lat_multicast(int ifn)
 {
     struct ifreq interface_ifreq;
     int dummy_fd;
+#ifdef HAVE_AF_LINK
+    struct sockaddr_dl *sockdl;
+#endif
+
+    memset(&interface_ifreq, 0, sizeof(interface_ifreq));
 
     /* if we don't have an interface, bail: */
     if (ifn != 0 || _latd_bpf_interface_name == NULL) {
@@ -600,17 +650,19 @@ int BPFInterfaces::remove_lat_multicast(int ifn)
       return -1;
     }
 
-    /* This is the LAT multicast address */
+#ifdef HAVE_AF_LINK
+    sockdl = (sockaddr_dl *)&interface_ifreq.ifr_addr;
+    sockdl->sdl_family = AF_LINK;
+    sockdl->sdl_alen = 6;
+    store_lat_multicast((unsigned char *)sockdl->sdl_data);
+#else
     interface_ifreq.ifr_addr.sa_family = AF_UNSPEC;
+    store_lat_multicast((unsigned char *)interface_ifreq.ifr_addr.sa_data);
+#endif
 #ifdef HAVE_SOCKADDR_SA_LEN
     interface_ifreq.ifr_addr.sa_len = sizeof(interface_ifreq.ifr_addr);
 #endif /* HAVE_SOCKADDR_SA_LEN */
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[0]  = 0x09;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[1]  = 0x00;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[2]  = 0x2b;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[3]  = 0x00;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[4]  = 0x00;
-    ((unsigned char *) interface_ifreq.ifr_addr.sa_data)[5]  = 0x0f;
+
     if (ioctl(dummy_fd, SIOCDELMULTI, &interface_ifreq) < 0) {
       debuglog(("bpf: failed to delete from the multicast list for interface for %s: %s\n",
 		interface_ifreq.ifr_name, strerror(errno)));
@@ -644,7 +696,7 @@ int BPFInterfaces::bind_socket(int ifn)
       syslog(LOG_ERR, "Can't create LAT protocol socket: %m\n");
       return -1;
     }
-    
+
     /* set the filter on the BPF device: */
     program.bf_len = sizeof(moprc_bpf_filter) / sizeof(moprc_bpf_filter[0]);
     program.bf_insns = moprc_bpf_filter;
